"""Renderable module.

This module provides the interface for creating objects in the EnSight session
that can be displayed via HTML over the websocket server interface.
"""
import hashlib
import os
import shutil
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union, no_type_check
import uuid
import warnings
import webbrowser

import requests

if TYPE_CHECKING:
    from ansys.pyensight.core import Session


def _get_ansysnexus_version(version: Union[int, str]) -> str:
    if int(version) < 242:
        return ""
    return str(version)


class Renderable:
    """Generates HTML pages for renderable entities.

    This class provides the underlying HTML remote webpage generation for
    the :func:`show<ansys.pyensight.core.Session.show>` method. The approach
    is to generate the renderable in the EnSight session and make the
    artifacts available via the websocket server. The artifacts are then
    wrapped with simple HTML pages, which are also served up by the websocket
    server. These HTML pages can then be used to populate iframes.

    Parameters
    ----------
    session :
        PyEnSight session to generate renderables for.
    cell_handle :
        Jupyter notebook cell handle (if any). The default is ``None``.
    width : int, optional
        Width of the renderable. The default is ``None``.
    height : int, optional
        Height of the renderable. The default is ``None``.
    temporal : bool, optional
        Whether to show data for all timesteps in an interactive
        WebGL-based browser viewer. The default is ``False``.
    aa : int, optional
        Number of antialiasing passes to use when rendering images.
        The default is ``1``.
    fps : float, optional
        Number of frames per second to use for animation playback. The
        default is ``30.0``.
    num_frames : int, optional
        Number of frames of static timestep to record for animation playback.
        The default is ``None``.

    """

    def __init__(
        self,
        session: "Session",
        cell_handle: Optional[Any] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        temporal: bool = False,
        aa: int = 1,
        fps: float = 30.0,
        num_frames: Optional[int] = None,
    ) -> None:
        self._session = session
        self._filename_index: int = 0
        self._guid: str = str(uuid.uuid1()).replace("-", "")
        self._download_names: List[str] = []
        # The Jupyter notebook cell handle (if any)
        self._cell_handle = cell_handle
        # the URL to the base HTML file for this entity
        self._url: Optional[str] = None
        # the pathname of the HTML file in the remote EnSight session
        self._url_remote_pathname: Optional[str] = None
        # the name passed to the pyensight session show() string
        self._rendertype: str = ""
        # Common attributes used by various subclasses
        self._width: Optional[int] = width
        self._height: Optional[int] = height
        self._temporal: bool = temporal
        self._aa: int = aa
        self._fps: float = fps
        self._num_frames: Optional[int] = num_frames
        #
        self._using_proxy = False
        # if we're talking directly to WS, then use 'http' otherwise 'https' for the proxy
        self._http_protocol = "http"
        try:
            if self._session.launcher._pim_instance is not None:
                self._using_proxy = True
                self._http_protocol = "https"
        except Exception:
            # the launcher may not be PIM aware; that's ok
            pass

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}( url='{self._url}' )"

    @no_type_check
    def _repr_pretty_(self, p: "pretty", cycle: bool) -> None:
        """Support the pretty module for better IPython support.

        Parameters
        ----------
        p : text, optional
            Pretty flag. The default is ``"pretty"``.
        cycle : bool, optional
            Cycle flag. The default is ``None``.

        """
        name = self.__class__.__name__
        p.text(f"{name}( url='{self._url}' )")

    def _get_query_parameters_str(self, params: Optional[Dict[str, str]] = None) -> str:
        """Generate any optional http query parameters.
        Return a string formatted as a URL query to tack on to the
        beginning part of the URL.  The string may be empty if there
        aren't any parameters.  The method takes a dict
        of parameters, possibly empty or None, and combines it with the
        parameters from the launcher, which also may be empty.
        """
        qp_dict = self._session.launcher._get_query_parameters()
        if qp_dict is None:
            # just in case
            qp_dict = {}
        if params:
            qp_dict.update(params)
        query_parameter_str = ""
        symbol = "?"
        for p in qp_dict.items():
            query_parameter_str += f"{symbol}{p[0]}={p[1]}"
            symbol = "&"
        return query_parameter_str

    def _generate_filename(self, suffix: str) -> Tuple[str, str]:
        """Create session-specific files and URLs.

        Every time this method is called, a new filename (on the EnSight host)
        and the associated URL for that file are generated. The caller
        provides the suffix for the names.

        Parameters
        ----------
        suffix: str
            Suffix of the file.

        Returns
        -------
        Tuple[str, str]
            Filename to use on the host system and the URL that accesses the
            file via REST calls to the websocket server.

        """
        filename = f"{self._session.secret_key}_{self._guid}_{self._filename_index}{suffix}"
        # Note: cannot use os.path.join here as the OS of the EnSight session might not match
        # the client OS.
        pathname = f"{self._session.launcher.session_directory}/{filename}"
        self._filename_index += 1
        return pathname, filename

    def _generate_url(self) -> None:
        """Build the remote HTML filename and associated URL.

        On the remote system the, pathname to the HTML file is
        ``{session_directory}/{session}_{guid}_{index}_{type}.html``.
        The URL to the file (through the session HTTP server) is
        ``http://{system}:{websocketserverhtmlport}/{session}_{guid}_{index}_{type}.html``.

        Note that there may be optional http query parameters at the end of the URL.

        After this call, ``_url`` and ``_url_remote_pathname`` reflect these names.

        """
        suffix = f"_{self._rendertype}.html"
        filename_index = self._filename_index
        remote_pathname, _ = self._generate_filename(suffix)
        simple_filename = f"{self._session.secret_key}_{self._guid}_{filename_index}{suffix}"
        url = f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}"
        self._url = f"{url}/{simple_filename}{self._get_query_parameters_str()}"
        self._url_remote_pathname = remote_pathname

    def _save_remote_html_page(self, html: str) -> None:
        """Create an HTML webpage on the remote host.

        Given a snippet of HTML, create a file on the remote server with the
        name generated by the ``_generate_url()`` method.
        The most common use is to generate an "iframe" wrapper around some HTML
        snippet.

        Parameters
        ----------
        html : str
            HTML snippet to wrap remotely.

        """
        # save "html" into a file on the remote server with filename .html
        cmd = f'open(r"""{self._url_remote_pathname}""", "w").write("""{html}""")'
        self._session.grpc.command(cmd, do_eval=False)

    def browser(self) -> None:
        """Open a web browser page to display the renderable content."""
        if self._url:
            webbrowser.open(self._url)

    @property
    def url(self) -> Optional[str]:
        """URL to the renderable content."""
        return self._url

    def _default_size(self, width: int, height: int) -> Tuple[int, int]:
        """Propose and return a size for a rectangle.

        The renderable may have been constructed with user-supplied width and height
        information. If so, that information is returned. If not, the width and
        height values passed to this method are returned.

        Parameters
        ----------
        width : int
            Width value to return if the renderable does not have a width.
        height : int
            Height value to return if the renderable does not have a height.

        Returns
        -------
        Tuple[int, int]
            Tuple (width, height) of the size values to use.

        """
        out_w = self._width
        if out_w is None:
            out_w = width
        out_h = self._height
        if out_h is None:
            out_h = height
        return out_w, out_h

    def update(self) -> None:
        """Update the visualization and display it.

        When this method is called, the graphics content is updated to the
        current EnSight instance state. For example, an image might be re-captured.
        The URL of the content stays the same, but the content that the URL displays is
        updated.

        If the renderable was created in the context of a Jupyter notebook cell,
        the original cell display is updated.

        """
        if self._cell_handle:
            from IPython.display import IFrame

            width, height = self._default_size(800, 600)
            self._cell_handle.update(IFrame(src=self._url, width=width, height=height))

    def delete(self) -> None:
        """Delete all server resources for the renderable.

        A renderable occupies resources in the EnSight :class:`Session<ansys.pyensight.core.Session>`
        instance. This method releases those resources. Once this method is called, the renderable
        can no longer be displayed.

        Notes
        -----
        This method has not yet been implemented.

        """
        pass

    def download(self, dirname: str) -> List[str]:
        """Download the content files for the renderable.

        A renderable saves files (such as images, mpegs, and geometry) in the EnSight instance.
        Normally, these files are accessed via the webpage specified in the URL property.
        This method allows for those files to be downloaded to a local directory so that they
        can be used for other purposes.

        .. note::
           Any previously existing files with the same name are overwritten.

        Parameters
        ----------
        dirname : str
            Name of the existing directory to save the files to.


        Returns
        -------
        list
            List of names for the downloaded files.

        Examples
        --------
        Download the PNG file generated by the image renderable.

        >>> img = session.show('image", width=640, height=480, aa=4)
        >>> names = img.download("/tmp")
        >>> png_pathname = os.path.join("/tmp", names[0])

        """
        for filename in self._download_names:
            url = f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}/{filename}{self._get_query_parameters_str()}"
            outpath = os.path.join(dirname, filename)
            with requests.get(url, stream=True) as r:
                with open(outpath, "wb") as f:
                    shutil.copyfileobj(r.raw, f)
        return self._download_names


class RenderableImage(Renderable):
    """Renders an image on the EnSight host system and makes it available via a webpage."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize RenderableImage."""
        super().__init__(*args, **kwargs)
        self._rendertype = "image"
        self._generate_url()
        # the HTML serves up a PNG file
        pathname, filename = self._generate_filename(".png")
        self._png_pathname = pathname
        self._png_filename = filename
        # the download is the png file
        self._download_names.append(self._png_filename)
        self.update()

    def update(self):
        """Update the image and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated
        as an iframe reference.

        """
        # save the image file on the remote host
        w, h = self._default_size(1920, 1080)
        cmd = f'ensight.render({w},{h},num_samples={self._aa}).save(r"""{self._png_pathname}""")'
        self._session.cmd(cmd)
        # generate HTML page with file references local to the websocket server root
        html = '<body style="margin:0px;padding:0px;">\n'
        html += f'<img src="/{self._png_filename}{self._get_query_parameters_str()}">\n'
        html += "</body>\n"
        # refresh the remote HTML
        self._save_remote_html_page(html)
        super().update()


class RenderableDeepPixel(Renderable):
    """Renders a deep pixel image on the EnSight host system and makes it available via a webpage."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._rendertype = "deep_pixel"
        self._generate_url()
        pathname, filename = self._generate_filename(".tif")
        self._tif_pathname = pathname
        self._tif_filename = filename
        # the download is the tiff file
        self._download_names.append(self._tif_filename)
        self.update()

    def update(self):
        """Update the deep pixel image and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as
        an iframe reference.
        """
        # save the (deep) image file
        # get the optional query parameters which may be an empty string
        # needed for proxy servers like ansys lab
        optional_query = self._get_query_parameters_str()
        w, h = self._default_size(1920, 1080)
        deep = f",num_samples={self._aa},enhanced=1"
        cmd = f'ensight.render({w},{h}{deep}).save(r"""{self._tif_pathname}""")'
        self._session.cmd(cmd)
        html_source = os.path.join(os.path.dirname(__file__), "deep_pixel_view.html")
        with open(html_source, "r") as fp:
            html = fp.read()
        # copy some files from Nexus
        cmd = "import shutil, enve, ceiversion, os.path\n"
        base_name = "os.path.join(enve.home(), f'nexus{ceiversion.nexus_suffix}', 'django', "
        base_name += "'website', 'static', 'website', 'scripts', "
        for script in ["geotiff.js", "geotiff_nexus.js", "bootstrap.min.js"]:
            name = base_name + f"'{script}')"
            cmd += f'shutil.copy({name}, r"""{self._session.launcher.session_directory}""")\n'
        name = "os.path.join(enve.home(), f'nexus{ceiversion.nexus_suffix}', 'django', "
        name += "'website', 'static', 'website', 'content', 'bootstrap.min.css')"
        cmd += f'shutil.copy({name}, r"""{self._session.launcher.session_directory}""")\n'
        self._session.cmd(cmd, do_eval=False)
        # With Bootstrap 5 (2025 R1), class names have '-bs-' in them, e.g. 'data-bs-toggle' vs 'data-toggle'
        bs_prefix = "bs-"
        jquery_version = ""
        if int(self._session._cei_suffix) < 251:
            bs_prefix = ""
            jquery_version = "-3.4.1"
        jquery = f"jquery{jquery_version}.min.js"
        cmd = "import shutil, enve, ceiversion, os.path\n"
        name = base_name + f"'{jquery}')"
        cmd += "try:"
        cmd += f'    shutil.copy({name}, r"""{self._session.launcher.session_directory}""")\n'
        cmd += "except Exception:"
        cmd += "    pass"
        name = "os.path.join(enve.home(), f'nexus{ceiversion.nexus_suffix}', 'django', "
        name += "'website', 'static', 'website', 'content', 'bootstrap.min.css')"
        cmd += f'shutil.copy({name}, r"""{self._session.launcher.session_directory}""")\n'
        self._session.cmd(cmd, do_eval=False)
        url = f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}"
        tiff_url = f"{url}/{self._tif_filename}{optional_query}"
        # replace some bits in the HTML
        html = html.replace("TIFF_URL", tiff_url)
        html = html.replace("ITEMID", self._guid)
        html = html.replace("OPTIONAL_QUERY", optional_query)
        html = html.replace("JQUERY_VERSION", jquery_version)
        html = html.replace("BS_PREFIX", bs_prefix)
        # refresh the remote HTML
        self._save_remote_html_page(html)
        super().update()


class RenderableMP4(Renderable):
    """Renders the timesteps of the current dataset into an MP4 file and displays the results."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._rendertype = "animation"
        self._generate_url()
        # the HTML serves up a PNG file
        pathname, filename = self._generate_filename(".mp4")
        self._mp4_pathname = pathname
        self._mp4_filename = filename
        # the download is the mp4 file
        self._download_names.append(self._mp4_filename)
        self.update()

    def update(self):
        """Update the animation and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as an
        iframe reference.

        """
        # save the image file on the remote host
        w, h = self._default_size(1920, 1080)
        # Assume this is a particle trace animation save...
        num_frames = self._num_frames
        st = 0
        if self._num_frames is None:
            # get the timestep limits, [0,0] is non-time varying
            st, en = self._session.ensight.objs.core.TIMESTEP_LIMITS
            num_frames = en - st + 1
        self._session.ensight.file.animation_rend_offscreen("ON")
        self._session.ensight.file.animation_screen_tiling(1, 1)
        self._session.ensight.file.animation_format("mpeg4")
        self._session.ensight.file.animation_format_options("Quality High Type 1")
        self._session.ensight.file.animation_frame_rate(self._fps)
        self._session.ensight.file.animation_rend_offscreen("ON")
        self._session.ensight.file.animation_numpasses(self._aa)
        self._session.ensight.file.animation_stereo("mono")
        self._session.ensight.file.animation_screen_tiling(1, 1)
        self._session.ensight.file.animation_file(self._mp4_pathname)
        self._session.ensight.file.animation_window_size("user_defined")
        self._session.ensight.file.animation_window_xy(w, h)
        self._session.ensight.file.animation_frames(num_frames)
        self._session.ensight.file.animation_start_number(st)
        self._session.ensight.file.animation_multiple_images("OFF")
        self._session.ensight.file.animation_raytrace_it("OFF")
        self._session.ensight.file.animation_raytrace_ext("OFF")
        self._session.ensight.file.animation_play_flipbook("OFF")
        self._session.ensight.file.animation_play_keyframe("OFF")

        if self._num_frames is None:
            # playing over time
            self._session.ensight.file.animation_play_time("ON")
            self._session.ensight.file.animation_reset_traces("OFF")
            self._session.ensight.file.animation_reset_time("ON")
        else:
            # recording particle traces/etc
            self._session.ensight.file.animation_play_time("OFF")
            self._session.ensight.file.animation_reset_traces("ON")
            self._session.ensight.file.animation_reset_time("OFF")

        self._session.ensight.file.animation_reset_flipbook("OFF")
        self._session.ensight.file.animation_reset_keyframe("OFF")
        self._session.ensight.file.save_animation()

        # generate HTML page with file references local to the websocket server root
        html = '<body style="margin:0px;padding:0px;">\n'
        html += f'<video width="{w}" height="{h}" controls>\n'
        html += f'    <source src="/{self._mp4_filename}{self._get_query_parameters_str()}" type="video/mp4" />\n'
        html += "</video>\n"
        html += "</body>\n"

        # refresh the remote HTML
        self._save_remote_html_page(html)
        super().update()


class RenderableWebGL(Renderable):
    """Renders an AVZ file (WebGL renderable) on the EnSight host system and makes it available via
    a webpage.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._rendertype = "webgl"
        self._generate_url()
        pathname, filename = self._generate_filename(".avz")
        self._avz_pathname = pathname
        self._avz_filename = filename
        # the download is the avz file
        self._download_names.append(self._avz_filename)
        self.update()

    def update(self):
        """Update the WebGL geometry and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as
        an iframe reference.
        """
        # save the .avz file
        self._session.ensight.part.select_all()
        self._session.ensight.savegeom.format("avz")
        # current timestep or all of the timesteps
        ts = self._session.ensight.objs.core.TIMESTEP
        st = ts
        en = ts
        if self._temporal:
            st, en = self._session.ensight.objs.core.TIMESTEP_LIMITS
        self._session.ensight.savegeom.begin_step(st)
        self._session.ensight.savegeom.end_step(en)
        self._session.ensight.savegeom.step_by(1)
        # Save the file
        self._session.ensight.savegeom.save_geometric_entities(self._avz_pathname)
        # generate HTML page with file references local to the websocket server root
        version = _get_ansysnexus_version(self._session._cei_suffix)
        if self._using_proxy:
            # if using pim we get the static content from the front end and not
            # where ensight is running, thus we use a specific URI host and not relative.
            html = f"<script src='{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}/ansys{version}/nexus/viewer-loader.js'></script>\n"
            html += f"<ansys-nexus-viewer src='{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}/{self._avz_filename}"
            html += f"{self._get_query_parameters_str()}'></ansys-nexus-viewer>\n"
        else:
            html = f"<script src='/ansys{version}/nexus/viewer-loader.js'></script>\n"
            html += f"<ansys-nexus-viewer src='/{self._avz_filename}{self._get_query_parameters_str()}'></ansys-nexus-viewer>\n"
        # refresh the remote HTML
        self._save_remote_html_page(html)
        super().update()


class RenderableVNC(Renderable):
    """Generates an ansys-nexus-viewer component that can be used to connect to the EnSight VNC remote image renderer."""

    def __init__(self, *args, **kwargs) -> None:
        ui = kwargs.get("ui")
        if kwargs.get("ui"):
            kwargs.pop("ui")
        super().__init__(*args, **kwargs)
        self._ui = ui
        self._generate_url()
        self._rendertype = "remote"
        self.update()

    def _update_2023R2_or_less(self):
        """Update the remote rendering widget and display it for
        backend EnSight of version earlier than 2024R1
        """
        query_params = {
            "autoconnect": "true",
            "host": self._session.html_hostname,
            "port": self._session.ws_port,
        }
        url = f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}"
        url += "/ansys/nexus/novnc/vnc_envision.html"
        url += self._get_query_parameters_str(query_params)
        self._url = url

    def update(self):
        """Update the remote rendering widget and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as an
        iframe reference.

        """
        optional_query = self._get_query_parameters_str()
        version = _get_ansysnexus_version(self._session._cei_suffix)
        ui = "simple" if not self._ui else self._ui
        if int(self._session._cei_suffix) < 242:  # pragma: no cover
            version = ""
            self._update_2023R2_or_less()  # pragma: no cover
        else:
            html = (
                f"<script src='/ansys{version}/nexus/viewer-loader.js{optional_query}'></script>\n"
            )
            rest_uri = (
                f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}"
            )
            ws_uri = (
                f"{self._http_protocol}://{self._session.html_hostname}:{self._session.ws_port}"
            )

            query_args = ""
            if self._using_proxy and optional_query:  # pragma: no cover
                query_args = f', "extra_query_args":"{optional_query[1:]}"'  # pragma: no cover

            attributes = ' renderer="envnc"'
            attributes += f" ui={ui}"
            attributes += ' active="true"'
            attributes += (
                " renderer_options='"
                + f'{{ "ws":"{ws_uri}", "http":"{rest_uri}", "security_token":"{self._session.secret_key}", "connect_to_running_ens":true {query_args} }}'
                + "'"
            )

            html += f"<ansys-nexus-viewer {attributes}></ansys-nexus-viewer>\n"

            # refresh the remote HTML
            self._save_remote_html_page(html)
        super().update()


# Undocumented class
class RenderableVNCAngular(Renderable):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._generate_url()
        self._rendertype = "remote"
        self.update()

    def update(self):
        optional_query = self._get_query_parameters_str()
        version = _get_ansysnexus_version(self._session._cei_suffix)
        base_content = f"""
<!doctype html>
<html lang="en" class="dark">
<head><base href="/ansys{version}/nexus/angular/">
  <meta charset="utf-8">
  <title>WebEnSight</title>
  <script src="/ansys{version}/nexus/viewer-loader.js"></script>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="icon" type="image/x-icon" href="ensight.ico">
<link rel="stylesheet" href="styles.css"></head>
<body>
"""
        module_with_attributes = "\n  <web-en-sight "
        module_with_attributes += f'wsPort="{self._session.ws_port}" '
        module_with_attributes += f'secretKey="{self._session.secret_key}"'
        if self._using_proxy and optional_query:  # pragma: no cover
            module_with_attributes += f' extraQueryArgs="{optional_query[1:]}"'
        module_with_attributes += ">\n"
        script_src = '<script src="runtime.js" type="module"></script><script src="polyfills.js" type="module"></script><script src="main.js" type="module"></script></body>\n</html>'
        content = base_content + module_with_attributes + script_src
        self._save_remote_html_page(content)
        super().update()


class RenderableEVSN(Renderable):
    """Generates a URL that can be used to connect to the EnVision VNC remote image renderer."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._rendertype = "remote_scene"
        self._generate_url()
        pathname, filename = self._generate_filename(".evsn")
        self._evsn_pathname = pathname
        self._evsn_filename = filename
        pathname, filename = self._generate_filename(".png")
        self._proxy_pathname = pathname
        self._proxy_filename = filename
        # the download is the evsn file
        self._download_names.append(self._evsn_filename)
        self.update()

    def update(self):
        """Update the remote rendering widget and display it.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as an
        iframe reference.

        """
        # Save the proxy image
        w, h = self._default_size(1920, 1080)
        cmd = f'ensight.render({w},{h},num_samples={self._aa}).save(r"""{self._proxy_pathname}""")'
        self._session.cmd(cmd)
        # save the .evsn file
        self._session.ensight.file.save_scenario_which_parts("all")
        self._session.ensight.file.scenario_format("envision")
        # current timestep or all of the timesteps
        if self._temporal:
            st, en = self._session.ensight.objs.core.TIMESTEP_LIMITS
            self._session.ensight.file.scenario_steptime_anim(1, st, en, 1.0)
        else:
            self._session.ensight.file.scenario_steptime_anim(0, 1, 1, 1)
        varlist = self._session.ensight.objs.core.VARIABLES.find(True, "ACTIVE")
        vars = [x.DESCRIPTION for x in varlist]
        self._session.ensight.variables.select_byname_begin(vars)
        # Save the file
        self._session.ensight.file.save_scenario_fileslct(self._evsn_pathname)

        # generate HTML page with file references local to the websocketserver root
        optional_query = self._get_query_parameters_str()
        version = _get_ansysnexus_version(self._session._cei_suffix)
        html = f"<script src='/ansys{version}/nexus/viewer-loader.js{optional_query}'></script>\n"
        server = f"{self._http_protocol}://{self._session.html_hostname}:{self._session.html_port}"

        # FIXME: This method doesn't work with Ansys Lab since the viewer seems to require
        # a full pathname to the file being generated by EnSight on a shared file system.
        # The following commented out line should replace the two after that, but that
        # prevents running locally from working since it's not using the full pathname to
        # the shared file. -MFK
        cleanname = self._evsn_filename.replace("\\", "/")
        attributes = f"src='{cleanname}'"
        # attributes = f"src='{cleanname}{optional_query}'"

        attributes += f" proxy_img='/{self._proxy_filename}{optional_query}'"
        attributes += " aspect_ratio='proxy'"
        attributes += " renderer='envnc'"
        http_uri = f'"http":"{server}"'
        ws_uri = (
            f'"ws":"{self._http_protocol}://{self._session.html_hostname}:{self._session.ws_port}"'
        )
        secrets = f'"security_token":"{self._session.secret_key}"'
        if not self._using_proxy or not optional_query:  # pragma: no cover
            attributes += f" renderer_options='{{ {http_uri}, {ws_uri}, {secrets} }}'"
        elif self._using_proxy and optional_query:  # pragma: no cover
            query_args = f'"extra_query_args":"{optional_query[1:]}"'  # pragma: no cover
            attributes += f" renderer_options='{{ {http_uri}, {ws_uri}, {secrets}, {query_args} }}'"  # pragma: no cover
        html += f"<ansys-nexus-viewer {attributes}></ansys-nexus-viewer>\n"
        # refresh the remote HTML
        self._save_remote_html_page(html)
        super().update()


class RenderableSGEO(Renderable):  # pragma: no cover
    """Generates a WebGL-based renderable that leverages the SGEO format/viewer interface for progressive geometry transport."""

    def __init__(self, *args, **kwargs) -> None:  # pragma: no cover
        super().__init__(*args, **kwargs)
        self._generate_url()
        pathname, filename = self._generate_filename("")
        # on the server, a JSON block can be accessed via:
        # {_sgeo_base_pathname}/geometry.sgeo
        # and the update files:
        # {_sgeo_base_pathname}/{names}.bin
        self._sgeo_base_pathname = pathname
        self._sgeo_base_filename = filename
        # Create the directory where the sgeo files will go '/{filename}/' URL base
        cmd = f'import os\nos.mkdir(r"""{self._sgeo_base_pathname}""")\n'
        self._session.cmd(cmd, do_eval=False)
        # get a stream ID
        self._stream_id = self._session.ensight.dsg_new_stream(sgeo=1)
        #
        self._revision = 0
        self.update()

    def update(self):  # pragma: no cover
        """Generate a SGEO geometry file.

        This method causes the EnSight session to generate an updated geometry SGEO
        file and content and then display the results in any attached WebGL viewer.

        If the renderable is part of a Jupyter notebook cell, that cell is updated as an
        iframe reference.

        """
        # Ask for an update to be generated
        remote_filename = f"{self._sgeo_base_pathname}/geometry.sgeo"
        self._session.ensight.dsg_save_update(
            remote_filename,
            urlprefix=f"/{self._sgeo_base_filename}/",
            stream=self._stream_id,
        )

        # Update the proxy image
        self._update_proxy()

        # If the first update, generate the HTML
        if self._revision == 0:
            # generate HTML page with file references local to the websocketserver root
            attributes = (
                f"src='/{self._sgeo_base_filename}/geometry.sgeo{self._get_query_parameters_str()}'"
            )
            attributes += f" proxy_img='/{self._sgeo_base_filename}/proxy.png{self._get_query_parameters_str()}'"
            attributes += " aspect_ratio='proxy'"
            attributes += " renderer='sgeo'"
            version = _get_ansysnexus_version(self._session._cei_suffix)
            html = f"<script src='/ansys{version}/nexus/viewer-loader.js{self._get_query_parameters_str()}'></script>\n"
            html += f"<ansys-nexus-viewer id='{self._guid}' {attributes}></ansys-nexus-viewer>\n"
            html += self._periodic_script()
            # refresh the remote HTML
            self._save_remote_html_page(html)
            # Subsequent updates are handled by the component itself
            super().update()

        # update the revision file
        rev_filename = f"{self._sgeo_base_pathname}/geometry.rev"
        cmd = f'with open(r"""{rev_filename}""", "w") as fp:\n'
        cmd += f'    fp.write("{self._revision}")\n'
        self._session.cmd(cmd, do_eval=False)

        self._revision += 1

    def _update_proxy(self):
        """Replace the current proxy image with the current view."""
        # save a proxy image
        w, h = self._default_size(1920, 1080)
        remote_filename = f"{self._sgeo_base_pathname}/proxy.png"
        cmd = f'ensight.render({w},{h},num_samples={self._aa}).save(r"""{remote_filename}""")'
        self._session.cmd(cmd, do_eval=False)

    def delete(self) -> None:
        try:
            _ = self._session.ensight.dsg_close_stream(self._stream_id)
        except Exception:
            pass
        super().delete()

    def _periodic_script(self) -> str:
        html_source = os.path.join(os.path.dirname(__file__), "sgeo_poll.html")
        with open(html_source, "r") as fp:
            html = fp.read()
        revision_uri = f"/{self._sgeo_base_filename}/geometry.rev{self._get_query_parameters_str()}"
        html = html.replace("REVURL_ITEMID", revision_uri)
        html = html.replace("ITEMID", self._guid)
        return html


class RenderableFluidsWebUI(Renderable):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._session.ensight_version_check("2025 R1")
        warnings.warn("The webUI is still under active development and should be considered beta.")
        self._rendertype = "webui"
        self._generate_url()
        self.update()

    def _generate_url(self) -> None:
        sha256_hash = hashlib.sha256()
        sha256_hash.update(self._session._secret_key.encode())
        token = sha256_hash.hexdigest()
        optional_query = self._get_query_parameters_str()
        port = self._session._webui_port
        if "instance_name" in self._session._launcher._get_query_parameters():
            instance_name = self._session._launcher._get_query_parameters()["instance_name"]
            # If using PIM, the port needs to be the 443 HTTPS Port;
            port = self._session.html_port
            # In the webUI code there's already a workflow to pass down the query parameter
            # ans_instance_id, just use it
            instance_name = self._session._launcher._get_query_parameters()["instance_name"]
            optional_query = f"?ans_instance_id={instance_name}"
        url = f"{self._http_protocol}://{self._session.html_hostname}:{port}"
        url += f"{optional_query}#{token}"
        self._url = url



"""
Python wrapper for the core enshellservice

This package defines the EnShellGRPC class which provides a simpler
interface to the EnShell gRPC interface.

Python binding for the EnShell gRPC API

This class provides an asynchronous interface to the EnShell
core gRPC interface.
"""
import logging
import os
import random
import re
import subprocess
import sys
from typing import Optional

from ansys.api.pyensight.v0 import enshell_pb2, enshell_pb2_grpc
from ansys.pyensight.core import DEFAULT_ANSYS_VERSION  # pylint: disable=import-outside-toplevel
import grpc


class EnShellGRPC(object):
    """Create an instance of the EnShell gRPC interface wrapper.

    The default is to make a connection to an EnShell gRPC server
    on port 12345 on the loopback host.  If requested to launch
    the server, it will be the current version.

    Parameters
    ----------
    port: int, optional
        The port number of the EnShell gRPC server
    host:
        The hostname of the EnShell gRPC server
    version:
        A specific EnShell version number to run (e.g. '232' for 2023R2)
    """

    def __init__(
        self, port: int = 12345, host: str = "127.0.0.1", version: str = DEFAULT_ANSYS_VERSION
    ):
        self._port = port
        self._host = host
        self._desired_version = version
        #
        self._pid = None
        self._channel = None
        self._stub = None
        #
        # self._security_token = str(random.randint(0, 1000000))
        self._security_token: Optional[int] = None
        #
        # values found from EnShell in the Container
        self._cei_home = None
        self._ansys_version = None

    def __del__(self):
        self.shutdown()

    def host(self):
        """Get the hostname for this connection.

        Returns
        -------
        str
            the current connection hostname.
        """
        return self._host

    def port(self):
        """Get the port number for this connection.

        Returns
        -------
        int
            The current connection port number.
        """
        return self._port

    def set_security_token(self, n: Optional[int] = None):
        """set the security token for the gRPC connection.

        EnShell supports a security token in either numeric (-security {int}) or
        string (ENSIGHT_SECURITY_TOKEN environmental variable) form.  If EnShell
        is using a security token, all gRPC calls must include this token.  This
        call sets the token for all rGPC calls made by this class.
        Note: for this module, the security token must be a in bytes() format.
        For example:  str(1000).encode("utf-8")

        Parameters
        ----------
        n : Optional[int], optional
            An string to be used as the security token, by default None
        """
        self._security_token = n  # pragma: no cover

    def set_random_security_token(self):
        """Set a random security token for the gRPC connection."""
        self._security_token = str(random.randint(0, 1000000))  # pragma: no cover

    def security_token(self):
        """Return the security token for the gRPC connection.

        Returns
        -------
        str
            Returns the current connection security token
        """
        return self._security_token

    def shutdown(self):
        """shut down all gRPC connections.

        If this class launched the EnShell client instance, it will
        send the gRPC exit() call and then shut down all connections.
        """
        # if we launched EnShell, shut it down.
        if self._pid is not None:  # pragma: no cover
            _ = self.stop_server()  # pragma: no cover

    def start_server(self):  # pragma: no cover
        """Start an EnShell gRPC server instance.

        If the host application wishes to launch an EnShell instance, start_server()
        will launch a batch mode EnShell application with the security token and
        a gRPC server started on the port passed in the constructor.
        """
        if self._pid is not None:
            return self._pid

        my_env = os.environ.copy()
        if self._desired_version != "":
            exe = f"enshell{self._desired_version}"
        else:
            exe = "enshell"
        cmd = [exe, "-app", "-grpc_server", str(self._port)]
        if self._security_token:
            cmd.append("-security")
            cmd.append(self._security_token)
        if sys.platform in ("win32", "cygwin"):
            cmd[0] += ".bat"
            # cmd.append("-minimize_console")
            # si = subprocess.STARTUPINFO()
            # si.dwFlags = subprocess.STARTF_USESHOWWINDOW | subprocess.CREATE_NEW_CONSOLE
            # si.wShowWindow = subprocess.SW_HIDE
            # f = subprocess.DETACHED_PROCESS
            # DETACHED_PROCESS = 0x00000008
            # self._pid = subprocess.Popen(cmd, creationflags=f, close_fds=True, env=my_env).pid
            # self._pid = subprocess.Popen(cmd, startupinfo=si, close_fds=True, env=my_env).pid
            logging.debug(f"command: {cmd}\n\n")
            self._pid = subprocess.Popen(cmd, close_fds=True, env=my_env).pid
        else:
            self._pid = subprocess.Popen(cmd, close_fds=True, env=my_env).pid
        return self._pid

    def stop_server(self):
        """Shut down any gPRC connection made by this class.

        First, if this class launched the EnShell instance, via start_server(), the
        exit_cleanly() gRPC command will be sent.  Second, the local gRPC connection is
        dropped
        """
        response = None
        # if we are connected and we started the server, we will emit the 'exit' message
        if self.is_connected():  # pragma: no cover
            response = self._stub.exit_cleanly(
                enshell_pb2.google_dot_protobuf_dot_empty__pb2.Empty(), metadata=self.metadata()
            )
        self._stub = None
        self._dsg_stub = None
        if self._channel:  # pragma: no cover
            self._channel.close()
        self._channel = None
        self._pid = None
        return response

    def is_connected(self):
        """Check if a gRPC connection has been established.

        Returns
        -------
        bool
            True if a previous connect() call made a valid gRPC connection.
        """
        if not self._channel:
            return False
        return self._channel is not None

    def connect(self, timeout: Optional[float] = 15.0):
        """Establish a connection to an EnShell gRPC server.

        Attempt to connect to an EnShell gRPC server using the host and port
        established by the constructor.  Note on failure, this function just
        returns, but is_connected() will return False.

        Parameters
        ----------
        timeout : Optional[float], optional
            timeout how long to wait for the connection to timeout., by default 15.0
        """
        if self._channel is not None:
            return
        self._channel = grpc.insecure_channel(
            "{}:{}".format(self._host, self._port),
            options=[
                ("grpc.max_receive_message_length", -1),
                ("grpc.max_send_message_length", -1),
                ("grpc.testing.fixed_reconnect_backoff_ms", 1100),
            ],
        )
        try:
            grpc.channel_ready_future(self._channel).result(timeout=timeout)
        except grpc.FutureTimeoutError:  # pragma: no cover
            self._channel = None  # pragma: no cover
            return  # pragma: no cover
        self._stub = enshell_pb2_grpc.EnShellServiceStub(self._channel)

    def connect_existing_channel(self, channel: grpc.Channel):  # pragma: no cover
        """Establish a connection to an EnShell gRPC server.

        Attempt to connect to an EnShell gRPC server using the host and port
        established by the constructor.  Note on failure, this function just
        returns, but is_connected() will return False.

        Parameters
        ----------
        channel : grpc.Channel
            Timeout how long to wait for the connection to timeout.
        """
        if self._channel is not None:
            raise RuntimeError("connect_existing_channel: channel already connected.")

        if channel is None:
            raise RuntimeError("connect_existing_channel: bad channel passed in.")

        self._channel = channel
        self._stub = enshell_pb2_grpc.EnShellServiceStub(self._channel)

    def metadata(self):  # pragma: no cover
        """Compute internal gRPC stream metadata."""
        ret = list()
        if self._security_token is not None:
            s = self._security_token
            if type(s) == str:
                s = s.encode("utf-8")
            ret.append((b"shared_secret", s))
        return ret

    def run_command(self, command_string: str):
        """send an EnShell command string to be executed in EnShell.

        The string will be sent to EnShell via the EnShellService::run_command()
        gRPC call.  An IOError exception may be thrown
        if there's a gRPC communication problem.  The response
        is the tuple of the EnShell return code and return string.

        Parameters
        ----------
        command_string : str
            Command_string the EnShell string to be executed.

        Returns
        -------
        tuple
            A tuple of (int, string) for (returnCode, returnString)
        """
        self.connect()
        if not self._stub:  # pragma: no cover
            return (0, "")  # pragma: no cover
        try:
            response = self._stub.run_command(
                enshell_pb2.EnShellCommandLine(command_line=command_string),
                metadata=self.metadata(),
            )
        except Exception:  # pragma: no cover
            raise IOError("gRPC connection dropped")  # pragma: no cover

        return (response.ret, response.response)

    def run_command_with_env(self, command_string: str, env_string: str):
        """send an EnShell command string and env var string to be executed in EnShell

        The string will be sent to EnShell via the EnShellService::run_command()
        gRPC call.  An IOError exception may be thrown
        if there's a gRPC communication problem.  The response
        is the tuple of the EnShell return code and return string.
        Parameters
        ----------
        command_string : str
            the EnShell string to be executed.
        env_string : str
            String of the environment.

        Returns
        -------
        Tuple
            A tuple of (int, string) for (returnCode, returnString)
        """
        self.connect()
        if not self._stub:  # pragma: no cover
            return (0, "")  # pragma: no cover
        try:
            response = self._stub.run_command_with_env(
                enshell_pb2.EnShellCommandWithEnvLine(
                    command_line=command_string, env_line=env_string
                ),
                metadata=self.metadata(),
            )
        except Exception:  # pragma: no cover
            raise IOError("gRPC connection dropped")  # pragma: no cover

        return (response.ret, response.response)

    # @brief Tell EnShell to start EnSight
    #
    # The string will be sent to EnShell via the EnShellService::run_command()
    # gRPC call.  An IOError exception may be thrown
    # if there's a gRPC communication problem.  The response
    # is the tuple of the EnShell return code and return string.
    # If ensight_env is used, the format is a single string of
    # environment variable name=value pairs with multiple pairs
    # separated by '\n' characters.
    # @param ensight_args arguments for the ensight command line
    # @param ensight_env optional environment variables to set before running EnSight
    # @return A tuple of (int, string) for (returnCode, returnString)
    def start_ensight(self, ensight_args: Optional[str] = None, ensight_env: Optional[str] = None):
        """Tell EnShell to start EnSight.

        The string will be sent to EnShell via the EnShellService::run_command()
        gRPC call.  An IOError exception may be thrown
        if there's a gRPC communication problem.  The response
        is the tuple of the EnShell return code and return string.

        Parameters
        ----------
        ensight_args : Optional[str], optional
            ensight_args arguments for the ensight command line, by default None

        Returns
        -------
        Tuple
            A tuple of (int, string) for (returnCode, returnString)
        """
        self.connect()

        command_string = "start_app CLIENT -c 127.0.0.1 -enshell"
        if ensight_args and (ensight_args != ""):
            command_string += " " + ensight_args

        if ensight_env is None or ensight_env == "":  # pragma: no cover
            return self.run_command(command_string)
        else:
            return self.run_command_with_env(command_string, ensight_env)  # pragma: no cover

    def start_ensight_server(
        self, ensight_args: Optional[str] = None, ensight_env: Optional[str] = None
    ):
        """Tell EnShell to start the EnSight server.

        The string will be sent to EnShell via the EnShellService::run_command()
        gRPC call.  An IOError exception may be thrown
        if there's a gRPC communication problem.  The response
        is the tuple of the EnShell return code and return string.

        Parameters
        ----------
        ensight_args : Optional[str], optional
            ensight_args arguments for the ensight command line, by default None

        Returns
        -------
        Tuple
            A tuple of (int, string) for (returnCode, returnString)
        """
        self.connect()
        command_string = (
            f"start_app OTHER /ansys_inc/v{self.ansys_version()}/CEI/bin/ensight_server"
        )
        if ensight_args and (ensight_args != ""):
            command_string += " " + ensight_args

        if ensight_env is None or ensight_env == "":  # pragma: no cover
            return self.run_command(command_string)
        else:
            return self.run_command_with_env(command_string, ensight_env)  # pragma: no cover

    # @brief
    #
    # @param cmd The command line
    # @return A tuple of (int, string) for (returnCode, returnString)
    def start_other(self, cmd: str, extra_env: Optional[str] = None):
        """Tell EnShell to start a non-EnShell aware command.

        The string will be sent to EnShell via the EnShellService::run_command()
        gRPC call.  An IOError exception may be thrown
        if there's a gRPC communication problem.  The response
        is the tuple of the EnShell return code and return string.

        Parameters
        ----------
        cmd : str
            _description_

        Returns
        -------
        _type_
            _description_
        """
        self.connect()
        command_string = "start_app OTHER " + cmd

        if extra_env is None or extra_env == "":  # pragma: no cover
            return self.run_command(command_string)
        else:
            return self.run_command_with_env(command_string, extra_env)  # pragma: no cover

    def cei_home(self):
        """Get the value of CEI_HOME from EnShell."""
        self._get_cei_home()
        return self._cei_home

    def ansys_version(self):
        """Get the Ansys version from EnShell (e.g. 232)"""
        self._get_cei_home()
        return self._ansys_version

    def _get_cei_home(self):
        if self._cei_home is not None:
            return

        self.connect()
        command_string = "show_ceihome"
        ret = self.run_command(command_string)
        # logging.debug(f"{command_string} :: ret = {ret}\n")
        if ret[0] != 0:  # pragma: no cover
            self._cei_home = None  # pragma: no cover
            raise RuntimeError("Error getting printenv from EnShell")  # pragma: no cover

        # split the newline delimited string into a list of strings
        env_vars = ret[1].strip().split("\n")
        # find the string containing CEI_HOME
        cei_home_line = [x for x in env_vars if "CEI_HOME" in x][0]
        if cei_home_line is None:  # pragma: no cover
            raise RuntimeError(
                "Error getting CEI_HOME env var from the Docker container.\n{ret}\n"
            )  # pragma: no cover

        # CEI_HOME is everything after the equal sign
        equal_sign_loc = cei_home_line.find("=")
        if equal_sign_loc < 0:  # pragma: no cover
            raise RuntimeError(
                "Error getting CEI_HOME env var from the Docker container.\n{ret}\n"
            )  # pragma: no cover
        self._cei_home = cei_home_line[equal_sign_loc + 1 :]
        m = re.search("/v(\d\d\d)/", self._cei_home)
        if not m:  # pragma: no cover
            self.stop_server()  # pragma: no cover
            raise RuntimeError(
                "Can't find version from cei_home in the Docker container.\n{ret}\n"
            )  # pragma: no cover
        self._ansys_version = m.group(1)



"""Docker Launcher module.

The Docker Launcher module provides PyEnSight with the ability to launch a
:class:`Session<ansys.pyensight.core.Session>` instance using a local Docker
installation or a `PIM <https://github.com/ansys/pypim`_-launched Docker installation.

Examples:
    ::

        from ansys.pyensight.core import DockerLauncher
        launcher = DockerLauncher(data_directory="D:\\data")
        launcher.pull()
        session = launcher.start()
        session.close()

"""
import logging
import os.path
import subprocess
from typing import TYPE_CHECKING, Any, Dict, Optional
import uuid

try:
    import grpc
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The grpc module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize grpc")

try:
    import docker
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The docker module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize Docker")

from ansys.pyensight.core.common import (
    find_unused_ports,
    get_file_service,
    launch_enshell_interface,
    populate_service_host_port,
    pull_image,
)
from ansys.pyensight.core.launcher import Launcher
import ansys.pyensight.core.session
from ansys.pyensight.core.session import Session

if TYPE_CHECKING:
    from docker import DockerClient
    from docker.models.containers import Container


class DockerLauncher(Launcher):
    """Creates a ``Session`` instance using a copy of EnSight hosted in a Docker container.

    This class allows you to either attach to a `PIM <https://github.com/ansys/pypim>`_-launched
    Docker container hosting EnSight or launch a local Docker container hosting EnSight.
    gRPC connections to EnShell and EnSight are established. A PyEnSight ``Session`` instance
    is returned upon successful launch and connection.

    Parameters
    ----------
    data_directory : str, optional
        Host directory to make into the container at ``/data``.
        The default is ``None``.
    docker_image_name : str, optional
        Name of the Docker image to use. The default is ``None``.
    use_dev : bool, optional
        Whether to use the latest ``ensight_dev`` Docker image. The
        default is ``False``. However, this value is overridden if the
        name of a Docker image is specified for the ``docker_image_name``
        parameter.
    channel :
        Existing gRPC channel to a running ``EnShell`` instance provided by PIM.
    pim_instance :
        PyPIM instance if using PIM (internal). The default is ``None``.
    timeout : float, optional
        Number of seconds to try a gRPC connection before giving up.
        This parameter is defined on the parent ``Launcher`` class,
        where the default is ``120``.
    use_egl : bool, optional
        Whether to use EGL hardware for accelerated graphics. The platform
        must be able to support this hardware. This parameter is defined on
        the parent ``Launcher`` class, where the default is ``False``.
    use_sos : int, optional
        Number of EnSight servers to use for SOS (Server of Server) mode.
        This parameter is defined on the parent ``Launcher`` class, where
        the default is ``None``, in which case SOS mode is not used.
    additional_command_line_options: list, optional
        Additional command line options to be used to launch EnSight.
        Arguments that contain spaces are not supported.

    Examples
    --------
    >>> from ansys.pyensight.core import DockerLauncher
    >>> launcher = DockerLauncher(data_directory="D:\\data")
    >>> launcher.pull()
    >>> session = launcher.start()
    >>> session.close()

    """

    def __init__(
        self,
        data_directory: Optional[str] = None,
        docker_image_name: Optional[str] = None,
        use_dev: bool = False,
        channel: Optional[grpc.Channel] = None,
        pim_instance: Optional[Any] = None,
        **kwargs,
    ) -> None:
        """Initialize DockerLauncher."""
        super().__init__(**kwargs)

        self._data_directory = data_directory
        self._enshell_grpc_channel = channel
        self._service_uris: Dict[Any, str] = {}
        self._image_name: Optional[str] = None
        self._docker_client: Optional["DockerClient"] = None
        self._container: Optional["Container"] = None
        self._enshell: Optional[Any] = None
        self._pim_instance: Optional[Any] = pim_instance

        # EnSight session secret key
        self._secret_key: str = str(uuid.uuid1())
        # temporary directory
        # it's in the ephemeral container, so just use "ensight's"
        # home directory within the container
        self._session_directory: str = "/home/ensight"
        # the Ansys / EnSight version we found in the container
        # to be reassigned later
        self._ansys_version: Optional[str] = None
        # EnShell's log file
        self._enshell_log_file = None
        # pointer to the file service object if available
        self._pim_file_service = None

        logging.debug(
            f"DockerLauncher data_dir={self._data_directory}\n"
            + f"  image_name={self._image_name}\n"
            + f"  use_dev={use_dev}\n"
        )

        if self._enshell_grpc_channel and self._pim_instance:
            service_set = ["grpc_private", "http", "ws"]
            # if self._launch_webui:
            #    service_set.append("webui")
            if not set(service_set).issubset(self._pim_instance.services):
                raise RuntimeError(
                    "If channel is specified, the PIM instance must have a list of length 3 "
                    + "containing the appropriate service URIs. It does not."
                )
            # grab the URIs for the 3 required services passed in from PIM
            self._service_host_port = populate_service_host_port(
                self._pim_instance, {}, webui=self._launch_webui
            )
            # attach to the file service if available
            self._pim_file_service = get_file_service(self._pim_instance)
            # if using PIM, we have a query parameter to append to http requests
            if self._pim_instance is not None:
                d = {"instance_name": self._pim_instance.name}
                self._add_query_parameters(d)
            #
            return

        # EnShell gRPC port, EnSight gRPC port, HTTP port, WSS port
        # skip 1999 as that internal to the container is used to the container for the VNC connection
        num_ports = 4
        if self._launch_webui:
            num_ports = 5
        ports = find_unused_ports(num_ports, avoid=[1999])
        if ports is None:  # pragma: no cover
            raise RuntimeError(
                "Unable to allocate local ports for EnSight session"
            )  # pragma: no cover
        self._service_host_port = {}
        self._service_host_port["grpc"] = ("127.0.0.1", ports[0])
        self._service_host_port["grpc_private"] = ("127.0.0.1", ports[1])
        self._service_host_port["http"] = ("127.0.0.1", ports[2])
        self._service_host_port["ws"] = ("127.0.0.1", ports[3])
        if self._launch_webui:
            self._service_host_port["webui"] = ("127.0.0.1", ports[4])

        # get the optional user-specified image name
        # Note: the default name needs to change over time...  TODO
        self._image_name = "ghcr.io/ansys-internal/ensight"
        if use_dev:
            self._image_name = "ghcr.io/ansys-internal/ensight_dev"
        if docker_image_name:
            self._image_name = docker_image_name

        # Load up Docker from the user's environment
        self._docker_client = docker.from_env()

    def ansys_version(self) -> Optional[str]:
        """Get the Ansys version (three-digit string) found in the Docker container.

        Returns
        -------
        str
            Ansys version or ``None`` if not found or not started.

        """
        return self._ansys_version

    def pull(self) -> None:
        """Pull the Docker image.

        Raises
        ------
        RuntimeError
            If the Docker image couldn't be pulled.

        """
        pull_image(self._docker_client, self._image_name)

    def _get_container_env(self) -> Dict:
        # Create the environmental variables
        # Environment to pass into the container
        container_env = {
            "ENSIGHT_SECURITY_TOKEN": self._secret_key,
            "WEBSOCKETSERVER_SECURITY_TOKEN": self._secret_key,
            "ENSIGHT_SESSION_TEMPDIR": self._session_directory,
        }

        # If for some reason, the ENSIGHT_ANSYS_LAUNCH is set previously,
        # honor that value, otherwise set it to "pyensight".  This allows
        # for an environmental setup to set the value to something else
        # (e.g. their "app").
        if "ENSIGHT_ANSYS_LAUNCH" not in os.environ:
            container_env["ENSIGHT_ANSYS_LAUNCH"] = "container"
        else:
            container_env["ENSIGHT_ANSYS_LAUNCH"] = os.environ["ENSIGHT_ANSYS_LAUNCH"]

        if self._pim_instance is None:
            container_env["ANSYSLMD_LICENSE_FILE"] = os.environ["ANSYSLMD_LICENSE_FILE"]
            if "ENSIGHT_ANSYS_APIP_CONFIG" in os.environ:
                container_env["ENSIGHT_ANSYS_APIP_CONFIG"] = os.environ["ENSIGHT_ANSYS_APIP_CONFIG"]

        if self._launch_webui:
            container_env["SIMBA_WEBSERVER_TOKEN"] = self._secret_key
            container_env["FLUENT_WEBSERVER_TOKEN"] = self._secret_key

        return container_env

    def start(self) -> "Session":
        """Start EnShell by running a local Docker EnSight image.

        Once this method starts EnShell, it connects over gRPC to EnShell running
        in the Docker container. Once connected, EnShell launches a copy of EnSight
        and WSS in the container. It then creates and binds a
        :class:`Session<ansys.pyensight.core.Session>` instance to the created
        EnSight gRPC connection and returns this instance.

        Returns
        -------
        object
            PyEnSight ``Session`` object instance.

        Raises
        ------
        RuntimeError
            Variety of error conditions.

        """
        tmp_session = super().start()
        if tmp_session:
            return tmp_session

        # Launch the EnSight Docker container locally as a detached container
        # initially running EnShell over the first gRPC port. Then launch EnSight
        # and other apps.

        # get the environment to pass to the container
        container_env = self._get_container_env()

        # Ports to map between the host and the container
        # If we're here in the code, then we're not using PIM
        # and we're not really using URIs where the hostname
        # is anything other than 127.0.0.1, so, we only need
        # to grab the port numbers.
        grpc_port = self._service_host_port["grpc"][1]
        ports_to_map = {
            str(self._service_host_port["grpc"][1])
            + "/tcp": str(self._service_host_port["grpc"][1]),
            str(self._service_host_port["grpc_private"][1])
            + "/tcp": str(self._service_host_port["grpc_private"][1]),
            str(self._service_host_port["http"][1])
            + "/tcp": str(self._service_host_port["http"][1]),
            str(self._service_host_port["ws"][1]) + "/tcp": str(self._service_host_port["ws"][1]),
        }
        if self._launch_webui:
            ports_to_map.update(
                {
                    str(self._service_host_port["webui"][1])
                    + "/tcp": str(self._service_host_port["webui"][1])
                }
            )
        # The data directory to map into the container
        data_volume = None
        if self._data_directory:
            data_volume = {self._data_directory: {"bind": "/data", "mode": "rw"}}

        # Start the container in detached mode with EnShell as a
        # gRPC server as the command
        #
        enshell_cmd = "-app -v 3 -grpc_server " + str(grpc_port)

        use_egl = self._use_egl()

        logging.debug("Starting Container...\n")
        if data_volume:
            if use_egl:  # pragma: no cover
                if self._docker_client:
                    self._container = self._docker_client.containers.run(  # pragma: no cover
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
            else:
                logging.debug(f"Running container {self._image_name} with cmd {enshell_cmd}\n")
                logging.debug(f"ports to map: {ports_to_map}\n")
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
                logging.debug(f"_container = {str(self._container)}\n")
        else:
            if use_egl:  # pragma: no cover
                if self._docker_client:  # pragma: no cover
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
            else:  # pragma: no cover
                logging.debug(
                    f"Running container {self._image_name} with cmd {enshell_cmd}\n"
                )  # pragma: no cover
                logging.debug(f"ports to map: {ports_to_map}\n")  # pragma: no cover
                if self._docker_client:  # pragma: no cover
                    self._container = self._docker_client.containers.run(  # pragma: no cover
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
                # logging.debug(f"_container = {str(self._container)}\n")
        logging.debug("Container started.\n")
        return self.connect()

    def launch_webui(self, container_env_str):
        # Run websocketserver
        cmd = f"cpython /ansys_inc/v{self._ansys_version}/CEI/"
        cmd += f"nexus{self._ansys_version}/nexus_launcher/webui_launcher.py"
        # websocket port - this needs to come first since we now have
        # --add_header as a optional arg that can take an arbitrary
        # number of optional headers.
        webui_port = self._service_host_port["webui"][1]
        grpc_port = self._service_host_port["grpc_private"][1]
        http_port = self._service_host_port["http"][1]
        ws_port = self._service_host_port["ws"][1]
        cmd += f" --server-listen-port {webui_port}"
        cmd += (
            f" --server-web-roots /ansys_inc/v{self._ansys_version}/CEI/nexus{self._ansys_version}/"
        )
        cmd += f"ansys{self._ansys_version}/ensight/WebUI/web/ui/"
        cmd += f" --ensight-grpc-port {grpc_port}"
        cmd += f" --ensight-html-port {http_port}"
        cmd += f" --ensight-ws-port {ws_port}"
        cmd += f" --ensight-session-directory {self._session_directory}"
        cmd += f" --ensight-secret-key {self._secret_key}"
        logging.debug(f"Starting WebUI: {cmd}\n")
        ret = self._enshell.start_other(cmd, extra_env=container_env_str)
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(f"Error starting WebUI: {cmd}\n")  # pragma: no cover

    def connect(self):
        """Create and bind a :class:`Session<ansys.pyensight.core.Session>` instance
        to the created EnSight gRPC connection started by EnShell.

        This method is an internal method.

        Returns
        -------
        obj
            :class:`Session<ansys.pyensight.core.Session>` instance

        Raises
        ------
        RuntimeError:
            Variety of error conditions.
        """
        #
        #
        # Start up the EnShell gRPC interface
        self._enshell = launch_enshell_interface(
            self._enshell_grpc_channel, self._service_host_port["grpc"][1], self._timeout
        )
        log_dir = "/data"
        if self._enshell_grpc_channel:  # pragma: no cover
            log_dir = "/work"

        if not self._enshell.is_connected():  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError("Can't connect to EnShell over gRPC.")  # pragma: no cover

        cmd = "set_no_reroute_log"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error sending EnShell command: {cmd} ret: {ret}")

        files_to_try = [log_dir + "/enshell.log", "/home/ensight/enshell.log"]
        for f in files_to_try:  # pragma: no cover
            cmd = "set_debug_log " + f
            ret = self._enshell.run_command(cmd)
            if ret[0] == 0:
                self._enshell_log_file = f
                break
            else:
                logging.debug(f"enshell error; cmd: {cmd} ret: {ret}\n")

        if self._enshell_log_file is not None:  # pragma: no cover
            logging.debug(f"enshell log file {self._enshell_log_file}\n")

        cmd = "verbose 3"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(
                f"Error sending EnShell command: {cmd} ret: {ret}"
            )  # pragma: no cover

        logging.debug("Connected to EnShell.  Getting CEI_HOME and Ansys version...\n")
        logging.debug(f"  _enshell: {self._enshell}\n\n")
        # Build up the command to run ensight via the EnShell gRPC interface

        self._cei_home = self._enshell.cei_home()
        self._ansys_version = self._enshell.ansys_version()
        print("CEI_HOME=", self._cei_home)
        print("Ansys Version=", self._ansys_version)

        logging.debug("Got them.  Starting EnSight...\n")

        use_egl = self._use_egl()

        # get the environment to pass to the container
        container_env_str = ""
        if self._pim_instance is not None:
            container_env = self._get_container_env()
            for i in container_env.items():
                container_env_str += f"{i[0]}={i[1]}\n"

        # Run EnSight
        ensight_env_vars = None
        if container_env_str != "":  # pragma: no cover
            ensight_env_vars = container_env_str  # pragma: no cover

        if use_egl:
            if ensight_env_vars is None:  # pragma: no cover
                ensight_env_vars = (
                    "LD_PRELOAD=/usr/local/lib64/libGL.so.1:/usr/local/lib64/libEGL.so.1"
                )
            else:
                ensight_env_vars += (  # pragma: no cover
                    "LD_PRELOAD=/usr/local/lib64/libGL.so.1:/usr/local/lib64/libEGL.so.1"
                )

        ensight_args = "-batch -v 3"

        if use_egl:
            ensight_args += " -egl"

        if self._use_sos:
            ensight_args += " -sos -nservers " + str(int(self._use_sos))

        ensight_args += " -grpc_server " + str(self._service_host_port["grpc_private"][1])

        vnc_url = "vnc://%%3Frfb_port=1999%%26use_auth=0"
        ensight_args += " -vnc " + vnc_url
        if self._additional_command_line_options:
            ensight_args += " "
            ensight_args += " ".join(self._additional_command_line_options)

        logging.debug(f"Starting EnSight with args: {ensight_args}\n")
        ret = self._enshell.start_ensight(ensight_args, ensight_env_vars)
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(
                f"Error starting EnSight with args: {ensight_args}"
            )  # pragma: no cover

        logging.debug("EnSight started.  Starting wss...\n")

        # Run websocketserver
        wss_cmd = "cpython /ansys_inc/v" + self._ansys_version + "/CEI/nexus"
        wss_cmd += self._ansys_version + "/nexus_launcher/websocketserver.py"
        # websocket port - this needs to come first since we now have
        # --add_header as a optional arg that can take an arbitrary
        # number of optional headers.
        wss_cmd += " " + str(self._service_host_port["ws"][1])
        #
        wss_cmd += " --http_directory " + self._session_directory
        # http port
        wss_cmd += " --http_port " + str(self._service_host_port["http"][1])
        # vnc port
        wss_cmd += " --client_port 1999"
        # optional PIM instance header
        if self._pim_instance is not None:
            # Add the PIM instance header. wss needs to return this optional
            # header in each http response.  It's how the Ansys Lab proxy
            # knows how to map back to this particular container's IP and port.
            wss_cmd += " --add_header instance_name=" + self._pim_instance.name
        # EnSight REST API
        if self._enable_rest_api:
            # grpc port
            wss_cmd += " --grpc_port " + str(self._service_host_port["grpc_private"][1])
        # EnVision sessions
        wss_cmd += " --local_session envision 5"

        wss_env_vars = None
        if container_env_str != "":  # pragma: no cover
            wss_env_vars = container_env_str  # pragma: no cover

        logging.debug(f"Starting WSS: {wss_cmd}\n")
        ret = self._enshell.start_other(wss_cmd, extra_env=wss_env_vars)
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(f"Error starting WSS: {wss_cmd}\n")  # pragma: no cover

        logging.debug("wss started.  Making session...\n")

        # build the session instance
        # WARNING: assuming the host is the same for grpc_private, http, and ws
        # This may not be true in the future if using PIM.
        # revise Session to handle three different hosts if necessary.
        use_sos = False
        if self._use_sos:
            use_sos = True
        if self._pim_instance is None:
            ws_port = self._service_host_port["ws"][1]
        else:
            ws_port = self._service_host_port["http"][1]  # pragma: no cover
        session = ansys.pyensight.core.session.Session(
            host=self._service_host_port["grpc_private"][0],
            grpc_port=self._service_host_port["grpc_private"][1],
            html_hostname=self._service_host_port["http"][0],
            html_port=self._service_host_port["http"][1],
            ws_port=ws_port,
            install_path=None,
            secret_key=self._secret_key,
            timeout=self._timeout,
            sos=use_sos,
            rest_api=self._enable_rest_api,
            webui_port=self._service_host_port["webui"][1] if self._launch_webui else None,
        )
        session.launcher = self
        self._sessions.append(session)

        if self._launch_webui:
            self.launch_webui(container_env_str)
        logging.debug("Return session.\n")

        return session

    def stop(self) -> None:
        """Release any additional resources allocated during launching."""
        if self._enshell:
            if self._enshell.is_connected():  # pragma: no cover
                try:
                    logging.debug("Stopping EnShell.\n")
                    self._enshell.stop_server()
                except Exception:  # pragma: no cover
                    pass  # pragma: no cover
                self._enshell = None
        #
        if self._container:
            try:
                logging.debug("Stopping the Docker Container.\n")
                self._container.stop()
            except Exception:
                pass
            try:
                logging.debug("Removing the Docker Container.\n")
                self._container.remove()
            except Exception:
                pass
            self._container = None

        if self._pim_instance is not None:
            logging.debug("Deleting the PIM instance.\n")
            self._pim_instance.delete()
            self._pim_instance = None
        super().stop()

    def file_service(self) -> Optional[Any]:
        """Get the PIM file service object if available."""
        return self._pim_file_service

    def _is_system_egl_capable(self) -> bool:
        """Check if the system is EGL capable.

        Returns
        -------
        bool
            ``True`` if the system is EGL capable, ``False`` otherwise.
        """
        if self._is_windows():
            return False  # pragma: no cover

        # FIXME: MFK, need to figure out how we'd do this
        # with a PIM managed system such as Ansys Lab
        if self._pim_instance is not None:
            return False

        try:  # pragma: no cover
            subprocess.check_output("nvidia-smi")
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False

    def enshell_log_contents(self) -> Optional[str]:
        """Get the contents of the EnShell log if possible.

        Returns
        -------
        str
           Contents of the log or ``None``.
        """
        if self._enshell_log_file is None:  # pragma: no cover
            return None  # pragma: no cover

        if self._container is not None:
            try:
                # docker containers allow copying from a container to the host
                # a file, files, or a directory.  it arrives as a tar file.
                # we're grabbing just the log file, but we still need to
                # extract it from the tar file and then put the contents
                # in the string we're returning
                from io import BytesIO
                import tarfile

                bits, stat = self._container.get_archive(self._enshell_log_file)
                file_obj = BytesIO()
                for chunk in bits:
                    file_obj.write(chunk)
                file_obj.seek(0)
                tar = tarfile.open(mode="r", fileobj=file_obj)
                member = tar.getmembers()
                text = tar.extractfile(member[0])
                s = text.read().decode("utf-8")
                text.close()
                return s
            except Exception as e:  # pragma: no cover
                logging.debug(f"Error getting EnShell log: {e}\n")  # pragma: no cover
                return None  # pragma: no cover

        fs = self.file_service()  # pragma: no cover
        if fs is not None:  # pragma: no cover
            try:
                fs.download_file("enshell.log", ".")
                f = open("enshell.log")
                s = f.read()
                f.close()
                return s
            except Exception:
                return None
